from google.cloud.datastore import Entity as DatastoreEntity, Key as DatastoreKey
from six import string_types
from echo.datastore.errors import InvalidKeyError, NotSavedException
from echo.datastore import properties, db_utils

# Override documentation for paths in this dictionary
__pdoc__ = {}


class BaseEntityMeta(type):
    __pdoc__['BaseEntityMeta'] = False

    # Needed to support setting property names on python2
    def __new__(mcls, name, bases, attrs):
        cls = super(BaseEntityMeta, mcls).__new__(mcls, name, bases, attrs)
        for attr, obj in attrs.items():
            if isinstance(obj, properties.Property):
                obj.__set_name__(cls, attr)
        return cls


class Entity(object):
    """Creates a datastore document under the entity [EntityName]

    Args:
        **data (kwargs): Values for properties in the new record, e.g User(name="Bob")

    Notes:
        Entities can be directly compared for equality each other e.g.
        ```entity.get(some_key) == entity.get(some_key)```
    """
    __metaclass__ = BaseEntityMeta

    def __init__(self, **data):
        if type(self) is Entity:
            raise Exception("You must extend Entity")
        project = db_utils.__client__().project
        self.__changes__ = []
        if "__entity__" in data:
            self.__datastore_entity__ = data.get("__entity__")
            del data["__entity__"]
        elif "id" in data:
            self.__datastore_entity__ = DatastoreEntity(key=Key(self.__entity_name__(), data.get("id"), project=project))
            del data["id"]
        else:
            self.__datastore_entity__ = DatastoreEntity(key=Key(self.__entity_name__(), project=project))

        for key, value in data.items():
            setattr(self, key, value)
        self.__field_set = self.__get_field_set()

    def __get_field_set(self):
        props = []
        for key, value in self.__class__.__dict__.items():
            try:
                if issubclass(value.__class__, properties.Property):
                    props.append((key, value.required, value.default))
            except TypeError:
                pass
        return props

    def key(self):
        """Generates a key for this Entity
        Returns:
            An instance of a key, convert to string to get a urlsafe key

        Raises:
            NotSavedException: Raised if reading a key of an unsaved entity unless the ID is
                explicitly provided
        """
        if self.__datastore_entity__.key.is_partial:
            raise NotSavedException()
        return Key(*self.__datastore_entity__.key.flat_path, project=db_utils.__client__().project)

    @classmethod
    def get(cls, key):
        """
        Get an entity with the specified key

        Args:
            key: A urlsafe key string or an instance of a Key

        Returns:
            An instance of the entity with the provided id

            Returns None if the id doesn't exist in the database

        Raises:
            InvalidKeyError: Raised if the key provided is invalid for this entity
        """
        if isinstance(key, string_types):
            try:
                key = Key.from_legacy_urlsafe(key)
            except Exception:
                raise InvalidKeyError(cls)

        if not isinstance(key, Key) or cls.__entity_name__() != key.kind:
            raise InvalidKeyError(cls)
        ds_entity = db_utils.__client__().get(key=key)
        if ds_entity:
            entity = cls(__entity__=ds_entity)
            return entity

    @classmethod
    def get_by_id(cls, entity_id):
        """
        Get an entity with a specified ID(Integer) or Name(String).

        Args:
            entity_id: An integer(id) or string(name) uniquely identifying the object

        Returns:
            An instance of the entity with the provided id

            Returns None if the id doesn't exist in the database
        """
        key = Key(cls.__name__, entity_id, project=db_utils.__client__().project)
        return cls.get(key)

    @classmethod
    def query(cls, limit=None, eventual=False, keys_only=False, order_by=None):
        """
        Filter entities based on certain conditions, an empty query will return all entities

        Args:
            limit: Maximum number of results to return, returns null by default
            eventual:
                Defaults to strongly consistent (False). Setting True will use eventual consistency,
                but cannot be used inside a transaction or will raise ValueError
            keys_only: Sets the results to include keys only
            order_by:
                A list of field names to order by, add `-` to order descending
                e.g. ["name", "-age"]

        Returns:
            A iterable query instance; call fetch() to get the results as a list.
        """
        return Query(cls, keys_only=keys_only, eventual=eventual, limit=limit, order_by=order_by)

    def is_saved(self):
        """Checks if an entity has any changes since read via get or query or last put.
        Always returns true for a new entity

        Returns:
            Boolean: True if no changes have been made
        """
        if self.__changes__ or self.__datastore_entity__.key.is_partial:
            return False
        return True

    def __pre_put_check__(self):
        for name, required, default in self.__field_set:
            if name not in self.__datastore_entity__:
                if default is None and required:
                    raise ValueError("Required field '%s' is not set for %s" % (name, self.__entity_name__()))
                if default is not None:
                    self.__datastore_entity__[name] = default

    def put(self):
        """Save changes made on this entity to datastore. Won't call datastore if no changes were made"""
        db_utils.put(self)

    def post_put(self, changes):
        """Override this function to run logic after saving the entity

        Args:
            changes (list): A list of fields that have been updated during put

        Notes:
            This function won't be called if there're no changes
        """

    def pre_put(self):
        """Override this function to run logic just before saving the entity
        NB: This function won't be called if no changes were made. i.e. when self.is_saved() == True
        """

    def delete(self):
        """Delete an entity from datastore"""
        db_utils.delete(self)

    def pre_delete(self):
        """Override this function to run any logic before deleting the entity. e.g. clear cache"""

    @classmethod
    def __entity_name__(cls):
        return cls.__name__

    def __eq__(self, other):
        return self.__datastore_entity__ == other.__datastore_entity__


class Query(object):
    def __init__(self, entity, keys_only=False, eventual=False, limit=None, order_by=None):
        order = []
        if isinstance(order_by, (list, tuple)):
            order = order_by
        elif isinstance(order_by, str):
            order = [order_by]
        self.__datastore_query = db_utils.__client__().query(kind=entity.__entity_name__(), order=order)
        self.entity = entity
        self.keys_only = keys_only
        if keys_only:
            self.__datastore_query.keys_only()
        self.limit = limit
        self.eventual = eventual
        self.__iterator = None

    def equal(self, field, value):
        """
        Equal to filter
        Args:
            field: Field name
            value: Value to compare

        Returns:
            Current Query Instance
        """
        self.__datastore_query.add_filter(field, '=', value)
        return self

    def gt(self, field, value):
        """
        Greater Than filter
        Args:
            field: Field name
            value: Value to compare

        Returns:
            Current Query Instance
        """
        self.__datastore_query.add_filter(field, '>', value)
        return self

    def gte(self, field, value):
        """
        Greater Than or Equal to filter
        Args:
            field: Field name
            value: Value to compare

        Returns:
            Current Query Instance
        """
        self.__datastore_query.add_filter(field, '>=', value)
        return self

    def lt(self, field, value):
        """
        Less Than filter
        Args:
            field: Field name
            value: Value to compare

        Returns:
            Current Query Instance
        """
        self.__datastore_query.add_filter(field, '<', value)
        return self

    def lte(self, field, value):
        """
        Less Than or Equal to filter
        Args:
            field: Field name
            value: Value to compare

        Returns:
            Current Query Instance
        """
        self.__datastore_query.add_filter(field, '<=', value)
        return self

    def fetch(self):
        """
        Get Query results as a list
        """
        return [entity for entity in self]

    def __process_result_item(self, result_item):
        if self.keys_only:
            # The result is a datastore entity with only a key
            return Key(result_item.kind, result_item.id, project=db_utils.__client__().project)
        entity = self.entity(__entity__=result_item)
        return entity

    def __iter__(self):
        return self

    def __next__(self):
        if not self.__iterator:
            self.__iterator = self.__datastore_query.fetch(limit=self.limit, eventual=self.eventual).__iter__()
        return self.__process_result_item(self.__iterator.__next__())

    __pdoc__["Query.next"] = False

    def next(self):
        # Support python2 iterators
        if not self.__iterator:
            self.__iterator = self.__datastore_query.fetch(limit=self.limit, eventual=self.eventual).__iter__()
        return self.__process_result_item(self.__iterator.next())


class Key(DatastoreKey):
    def __repr__(self):
        return self.to_legacy_urlsafe().decode('utf-8')

    def __str__(self):
        return self.__repr__()



from echo.datastore.entity import Entity
from echo.datastore.properties import TextProperty, DateTimeProperty, IntegerProperty
from echo.datastore.db_utils import put, put_async, delete, delete_async



class InvalidValueError(ValueError):
    """Raised if the value of a property does not fit the property type"""
    def __init__(self, _property, value):
        self.property = _property
        self.value = value

    def __str__(self):
        return "%s is not a valid value for property %s of type %s" % \
               (self.value, self.property.name, type(self.property).__name__)


class NotSavedException(Exception):
    """Raised when a key of an unsaved model is accessed"""
    def __str__(self):
        return "You can't read a key of an unsaved entity"


class InvalidKeyError(ValueError):
    """Raised when an invalid key is provided to the entity get method"""
    def __init__(self, entity):
        self.entity = entity

    def __str__(self):
        return "Invalid key for entity %s" % self.entity.__entity_name__()



from concurrent.futures import ThreadPoolExecutor
from future import builtins
from google.cloud.datastore.client import Client


def put(entities):
    """Save entities to datastore

    Args:
        entities: An entity or a list of entities
    """
    return put_async(entities).result()


def put_async(entities):
    """Save to datastore asynchronously
    Args:
        entities: An entity or a list of entities
    Returns:
        A future object. Use future.result() to wait for or get the results if ready
    """
    if not hasattr(entities, "__iter__"):
        entities = [entities]
    return __executor().submit(__process_and_put_entities, entities)


def delete(entities):
    return delete_async(entities).result()


def delete_async(entities):
    if not hasattr(entities, "__iter__"):
        entities = [entities]
    return __executor().submit(__process_and_delete_entities, entities)


def __process_and_put_entities(entities):
    datastore_entities = []
    changed_entities = []
    for entity in entities:
        entity.pre_put()
        entity.__pre_put_check__()
        if entity.is_saved():
            continue
        datastore_entities.append(entity.__datastore_entity__)
        changed_entities.append(entity)
    if datastore_entities:
        __client__().put_multi(datastore_entities)
        for entity in changed_entities:
            entity.post_put(entity.__changes__)
            entity.__changes__ = []


def __process_and_delete_entities(entities):
    datastore_keys = []
    for entity in entities:
        entity.pre_delete()
        datastore_keys.append(entity.key())
    __client__().delete_multi(datastore_keys)


def __executor():
    if not hasattr(builtins, "__datastore_executor__"):
        setattr(builtins, "__datastore_executor__", ThreadPoolExecutor(max_workers=2))
    return getattr(builtins, "__datastore_executor__")


def __client__():
    if not hasattr(builtins, "__datastore_client__"):
        setattr(builtins, "__datastore_client__", Client())
    return getattr(builtins, "__datastore_client__")



from echo.datastore.errors import InvalidValueError
from datetime import datetime
import pytz


class Property(object):
    """
    A class describing a typed, persisted attribute of a datastore entity
    """
    def __init__(self, default=None, required=False):
        """
        Args:
            default: The default value of the property
            required: Enforce the property value to be provided
        """
        self.default = None if default is None else self.validate(default)
        self.required = required
        self.name = None

    def __set_name__(self, owner, name):
        self.name = name

    def __set__(self, instance, value):
        value = self.validate(value)
        current_value = instance.__datastore_entity__.get(self.name)
        if current_value != value or self.name not in instance.__datastore_entity__:
            if self.name not in instance.__changes__:
                instance.__changes__.append(self.name)
            instance.__datastore_entity__[self.name] = value

    def __get__(self, instance, owner):
        value = instance.__datastore_entity__.get(self.name)
        # Set the default value if no value is written
        if (self.default is not None) and (self.name not in instance.__datastore_entity__):
            value = self.default
        value = self.user_value(value)
        return value

    def __delete__(self, instance):
        del instance.__datastore_entity__[self.name]
        if self.name not in instance.__changes__:
            instance.__changes__.append(self.name)

    def __type_check__(self, user_value, data_types):
        """
        Check whether this value has the right data type
        Args:
            user_value: The user_value you want to confirm
            data_types: Type/Types to check against

        Returns:
            user_value: A type checked user value or the default value
        """
        if not isinstance(user_value, data_types) and user_value is not None:
            raise InvalidValueError(self, user_value)
        return user_value

    def validate(self, user_value):
        """Validates the value provided by the user and converts it to a value acceptable to the database"""
        raise NotImplementedError("A property must implement `validate` function")

    def user_value(self, value):
        """Converts the database value to a value usable by the user"""
        raise NotImplementedError("A property must implement `user_value` function")


class TextProperty(Property):
    def validate(self, user_value):
        return self.__type_check__(user_value, str)

    def user_value(self, value):
        return value


class IntegerProperty(Property):
    def validate(self, user_value):
        return self.__type_check__(user_value, int)

    def user_value(self, value):
        return value


class DateTimeProperty(Property):
    """Accepts a python datetime instance
    Notes:
        - Dates are automatically localized to UTC
    """
    def __init__(self, auto_now_add=False, required=False):
        default = datetime.now() if auto_now_add else None
        super(DateTimeProperty, self).__init__(default=default, required=required)

    def validate(self, user_value):
        user_value = self.__type_check__(user_value, datetime)
        if user_value and not user_value.tzinfo:
            user_value = pytz.utc.localize(user_value)
        return user_value

    def user_value(self, value):
        return value



import setuptools


with open("requirements.txt", "r") as requirements:
    install_requires = requirements.read().strip().split("\n")

with open("README.md", "r") as readme:
    long_description = readme.read()

with open("RELEASE", "r") as release:
    version = release.read().strip()

setuptools.setup(
    name="echo-datastore-entities",
    version=version,
    author="Echo Mobile",
    author_email="engineering@echomobile.io",
    description="Client library for use with Google Cloud Datastore to provide an entities concept",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/echomobi/echo-datastore-entities",
    packages=['echo.datastore'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires
)
